--[[
#include "umf.lua"
]] --
local config = OptionsKeys {
	legacy = { "boolean", false },
	rotspeed = { "number", 1 },
	key = { rotate = { "string", "interact" }, throw = { "string", "r" }, unfreeze = { "string", "r" } },
}

local DEBUG = false

if DEBUG then
	PRINTTOSCREEN = true
end

local TOOL = {}

TOOL.printname = "Physics Gun"
TOOL.group = 6
TOOL.ammo = 10000

function TOOL:Initialize()
	self.mode = false
	self.scrolllock = 0
	self.motor_fast = LoadLoop( "MOD/sound/motor_fast.ogg" )
	self.motor_med = LoadLoop( "MOD/sound/motor_med.ogg" )

	self.world = {}
	if GetWorldBody then
		self.world[GetWorldBody()] = true
	end
	local allBodies = FindBodies( nil, true )
	local n, max_size = 0, 0
	for i, b in ipairs( allBodies ) do
		if not IsBodyDynamic( b ) then
			local min, max = GetBodyBounds( b )
			local size = VecLength( VecSub( max, min ) )
			if size > max_size and size < 1000000000 then
				n = i
				max_size = size
			end
		end
	end
	self.world[allBodies[n] or -1] = true
end

local function clamp( n, min, max )
	return math.min( math.max( n, min ), max )
end
local function steps( x, bounce )
	return x - math.abs( math.sin( x * math.pi ) / (math.pi - bounce) )
end
local function approach( c, t, m, mn )
	return c + math.max( math.min( t - c, m ), mn or -m )
end

local toolpos = Transform( Vec( 8.5 / 20, -7.5 / 20, -6 / 20 - 0.25 ), QuatEuler( 0, 0, 0 ) )
function TOOL:Update()
	SetToolTransform( toolpos )
end
local deploy_time = 0
local anim_data = { core = { { 0, 0 }, { 0, 0 }, { 0, 0 } }, arms = 0, rot = { 0, 0, 0 } }
local speedmod = { 1, 1.1, 1.05 }
function TOOL:Animate( body, shapes )
	SetToolTransform( toolpos )
	local armature = self.armature
	local dt = GetTimeStep()

	local arm_base_offset = 0
	local arm_tip_offset = 0
	if deploy_time < 3 then
		local stepf = steps( deploy_time * 2.7, 1.5 )

		local a, b = clamp( stepf, 0, 1 ), clamp( stepf, 1, 2 ) - 1
		local deploy = steps( clamp( 1 - deploy_time * 3, 0, 1 ), 0.8 )

		armature:SetBoneTransform( "root", Transform( Vec( 0, deploy * -0.2, 0 ), QuatEuler( deploy * -20, 0, 0 ) ) )

		armature:SetBoneTransform( "spine1", Transform( Vec( 0, 0, 0.1 - a / 10 ), QuatEuler( 0, 0, 0 ) ) )
		armature:SetBoneTransform( "spine2", Transform( Vec( 0, 0, 0.1 - b / 10 ), QuatEuler( 0, 0, 0 ) ) )

		arm_base_offset = 50 * a - 50
		arm_tip_offset = 90 * b - 90
	end
	local ang = self.grabbing and -20 or self.grabbed and -90 or -40
	local rand = self.grabbing and 6 or self.grabbed and 4 or 0
	anim_data.arms = approach( anim_data.arms, ang, 720 * dt )
	local a1, a2, a3 = anim_data.arms + (math.random() - 0.5) * rand, anim_data.arms + (math.random() - 0.5) * rand,
	                   anim_data.arms + (math.random() - 0.5) * rand

	if self.grabbing then
		anim_data.rot[2] = anim_data.rot[2] + dt * 200
		anim_data.rot[3] = GetTime() + 1
	elseif self.grabbed then
		anim_data.rot[2] = anim_data.rot[2] - dt * 100
		anim_data.rot[3] = GetTime() + 1
	elseif anim_data.rot[3] < GetTime() then
		local rt = (1 - math.random() ^ 3)
		anim_data.rot[3] = GetTime() + rt * 2
		anim_data.rot[2] = anim_data.rot[2] + (1 - rt) * (math.random() * 720 - 360)
	end
	if math.abs( anim_data.rot[1] - anim_data.rot[2] ) > 0.001 then
		PlayLoop( self.grabbed and self.motor_med or self.motor_fast, body:GetTransform().pos,
		          self.grabbed and 0.3 or self.grabbing and 0.2 or 1 )
	end
	anim_data.rot[1] = approach( anim_data.rot[1], anim_data.rot[2], 720 * dt )
	armature:SetBoneTransform( "arms_rot", Transform( Vec(), QuatEuler( 0, 0, anim_data.rot[1] ) ) )

	armature:SetBoneTransform( "arm0_base", Transform( Vec(), QuatEuler( a1 + arm_base_offset, 0, 0 ) ) )
	armature:SetBoneTransform( "arm0_tip", Transform( Vec(), QuatEuler( a2 + arm_tip_offset, 0, 0 ) ) )

	armature:SetBoneTransform( "arm1_base", Transform( Vec(), QuatEuler( a2 + arm_base_offset, 0, 0 ) ) )
	armature:SetBoneTransform( "arm1_tip", Transform( Vec(), QuatEuler( a3 + arm_tip_offset, 0, 0 ) ) )

	armature:SetBoneTransform( "arm2_base", Transform( Vec(), QuatEuler( a3 + arm_base_offset, 0, 0 ) ) )
	armature:SetBoneTransform( "arm2_tip", Transform( Vec(), QuatEuler( a1 + arm_tip_offset, 0, 0 ) ) )

	local warmth = math.min( deploy_time, 10 ) / 10
	local speed = self.grabbing and 360 or self.grabbed and 520 or 120
	for i = 1, 3 do
		if deploy_time > i / 2 then
			anim_data.core[i][1] = approach( anim_data.core[i][1], speed * speedmod[i], warmth * 500 * dt, -200 * dt )
			anim_data.core[i][2] = anim_data.core[i][2] + anim_data.core[i][1] * dt * ((i % 2) * 2 - 1)
		end
	end
	armature:SetBoneTransform( "core0", Transform( Vec( 0, 0, 0 ), QuatEuler( 0, 0, anim_data.core[1][2] ) ) )
	armature:SetBoneTransform( "core1", Transform( Vec( 0, 0, 0 ), QuatEuler( 0, 0, anim_data.core[2][2] ) ) )
	armature:SetBoneTransform( "core2", Transform( Vec( 0, 0, 0 ), QuatEuler( 0, 0, anim_data.core[3][2] ) ) )

	local lit = (anim_data.core[1][1] / 520) ^ 2
	for i = 1, #shapes do
		shapes[i]:SetEmissiveScale( lit )
	end

	deploy_time = deploy_time + 1.5 * dt

	if DEBUG then
		self:DrawDebug()
	end
end

function TOOL:Deploy()
	deploy_time = 0
	anim_data = { core = { { 0, 0 }, { 0, 0 }, { 0, 0 } }, arms = 0, rot = { 0, 0, GetTime() + 1.5 } }
end

function TOOL:GetTarget()
	local ray = PLAYER:GetCamera():Raycast( 100, -1 )
	if not IsValid( ray.shape ) then
		return
	end
	local body = ray.shape:GetBody()

	if self.world[GetEntityHandle( body )] or ray.shape:HasTag( "ground" ) or body:HasTag( "ground" ) then
		return
	end

	return body, ray.hitpos, ray.dist
end

function TOOL:AttemptGrab()
	local body, hitpos, dist = self:GetTarget()
	if not body then
		return
	end

	self.grabbed = body
	self.grabbed:SetDynamic( true )
	self.grabbed:SetVelocity( VEC_ZERO )
	local grtr = self.grabbed:GetTransform()
	self.relative = grtr:ToLocal( hitpos )
	self.rotation = PLAYER:GetTransform():ToLocal( self.grabbed:GetTransform() )
	self.dist = dist
	return self.grabbed
end

function TOOL:LeftClick()
	if not self:AttemptGrab() then
		self.grabbing = true
	end
end

function TOOL:RightClick()
	if self.grabbed then
		self.effect = { type = "freeze", body = self.grabbed, start = GetTime(), time = 1 }
		self.grabbed:SetVelocity( VEC_ZERO )
		self.grabbed:SetAngularVelocity( VEC_ZERO )
		self.grabbed:SetDynamic( false )
		self.grabbed = nil
		self.scrolllock = GetTime()
	else
		self.mode = not self.mode
	end
end

function TOOL:LeftClickReleased()
	self.grabbed = nil
	self.grabbing = false
	self.startrot = nil
end

function TOOL:MouseWheel( ds )
	if self.grabbed then
		self.dist = math.max( self.dist + ds, 2 )
		self.scrolllock = GetTime()
	end
end

function TOOL:ShouldLockMouseWheel()
	if self.grabbed or self.scrolllock > GetTime() - 0.5 then
		return true
	end
end

function TOOL:DrawBeam( source, target, object, r, g, b )
	if target == object then
		return DrawLine( source, target, r, g, b )
	end
	local prev = source
	for i = 1, 7 do
		local t = i / 8
		local t2 = t ^ 1.3
		local newpoint = source:Lerp( target, t2 ):Lerp( source:Lerp( object, t2 ), t )
		DrawLine( prev, newpoint, r, g, b )
		prev = newpoint
	end
	DrawLine( prev, object, r, g, b )
end

local effects = {
	freeze = function( body, t )
		body:DrawHighlight( .5 - t / 2 )
		body:DrawOutline( t, .5 + t / 2, 1, 1 - t )
	end,
	unfreeze = function( body, t )
		body:DrawHighlight( .5 - t / 2 )
		body:DrawOutline( 1, .5 + t / 2, t, 1 - t )
	end,
}

function TOOL:Tick()
	if self.effect then
		local t = (GetTime() - self.effect.start) / self.effect.time
		if t > 1 or not self.effect.body:IsValid() then
			self.effect = nil
		else
			local f = effects[self.effect.type]
			if f then
				f( self.effect.body, t )
			end
		end
	end
	local r, g, b = math.random(), math.random(), math.random()
	local camtr = PLAYER:GetCamera()
	local nozzle = self:GetBoneGlobalTransform( "nozzle" ).pos

	if not self.grabbed or not self.grabbed:IsValid() then
		if InputPressed( config.key.unfreeze ) then
			local body = self:GetTarget()
			if body then
				if not body:IsDynamic() then
					self.effect = { type = "unfreeze", body = body, start = GetTime(), time = 1 }
				end
				body:SetDynamic( true )
				body:SetVelocity( Vector( 0, 0, 0 ) )
			end
		end

		if self.grabbing then
			if self:AttemptGrab() then
				self.grabbing = false
				return
			end
			local ray = camtr:Raycast( 100, -1 )
			self:DrawBeam( nozzle, ray.hitpos, ray.hitpos, r, g, b )
			PointLight( nozzle, r, g, b, .1 )
			if not InputDown( "lmb" ) then
				self:LeftClickReleased()
			end
		end
		return
	end

	if InputPressed( config.key.throw ) then
		self.grabbed:SetVelocity( camtr.rot:Forward() * -100 )
		self.grabbed = nil
		return
	end

	self.grabbed:DrawOutline( r, g, b, 1 )
	local bodytr = self.grabbed:GetTransform()
	local onbody = bodytr:ToGlobal( self.relative )
	local aimpos = camtr:ToGlobal( Vector( 0, 0, -self.dist ) )

	self:DrawBeam( nozzle, aimpos, onbody, r, g, b )
	PointLight( nozzle, r, g, b, .2 )

	if DEBUG then
		DebugCross( onbody, 1, 0, 0 )
		DebugCross( aimpos, 0, 0, 1 )
		DebugLine( onbody, aimpos, 0, 1, 0 )
		DebugLine( nozzle, onbody, 1, 0, 0 )
		DebugLine( nozzle, aimpos, 0, 0, 1 )
	end

	local force = (aimpos - onbody) * 15
	self.grabbed:SetVelocity( force )
	-- ConstrainPosition(self.grabbed.handle, 0, onbody, aimpos)

	if InputDown( config.key.rotate ) then
		if not self.startrot then
			self.startrot = self.rotation.rot
			self.mousex, self.mousey = 0, 0
			self.playerlock = GetPlayerTransform( true )
		end
		if config.legacy then
			UiMakeInteractive()
			local dx, dy = UiGetMousePos()
			local w, h = UiWidth(), UiHeight()
			dx, dy = dx - w / 2, dy - h / 2
			self.rotation.rot = QuatEuler( (dy - self.mousey) / h * 360, (dx - self.mousex) / w * 360, 0 ) * self.rotation.rot
			self.mousex, self.mousey = dx, dy
		else
			SetPlayerTransform( self.playerlock, true )
			local s = config.rotspeed / 6
			self.rotation.rot = QuatEuler( s * InputValue( "mousedy" ), s * InputValue( "mousedx" ), 0 ) * self.rotation.rot
		end
		if InputPressed( "rmb" ) then
			self:RightClick()
			return
		end
	else
		self.startrot = nil
	end

	local nrot = PLAYER:GetTransform():ToGlobal( self.rotation )
	if InputDown( "shift" ) and self.startrot then
		local lp, ly, lr = MakeQuaternion( nrot.rot ):ToEuler()
		lp = math.floor( lp / 45 + .5 ) * 45
		ly = math.floor( ly / 45 + .5 ) * 45
		lr = math.floor( lr / 45 + .5 ) * 45
		nrot.rot = MakeQuaternion( QuatEuler( lp, ly, lr ) )
	end
	local p, y, r = (bodytr.rot * nrot.rot:Conjugate()):ToEuler()
	local diff = -Vector( p, y, r )

	self.grabbed:SetAngularVelocity( diff / 5 )
	-- ConstrainOrientation(self.grabbed.handle, 0, bodytr.rot, nrot.rot)

	if not InputDown( "lmb" ) or GetPlayerVehicle() ~= 0 then
		self:LeftClickReleased()
	end
end

TOOL.model = {
	prefab = [[
		<prefab version="0.9.2">
            <group name="instance=MOD/physgun.xml" pos="-3.5 0.9 9.1" rot="0.0 0.0 0.0">
                <vox pos="-0.125 -0.125 0.125" file="MOD/vox/physgun.vox" object="body" scale="0.5"/>
                <group name="spine0" pos="0.0 0.0 -0.025">
                    <group name="core0" pos="0.0 0.0 0.0" rot="0.0 0.0 0.0">
                        <vox pos="-0.025 -0.125 -0.05" rot="0.0 0.0 0.0" file="MOD/vox/physgun.vox" object="core_0" scale="0.5"/>
                    </group>
                    <group name="spine1" pos="0.0 0.0 -0.1">
                        <group name="core1" pos="0.0 0.0 0.0" rot="0.0 0.0 0.0">
                            <vox pos="-0.025 -0.125 -0.05" rot="0.0 0.0 0.0" file="MOD/vox/physgun.vox" object="core_1" scale="0.5"/>
                        </group>
                        <group name="spine2" pos="0.0 0.0 -0.1">
                            <group name="core2" pos="0.0 0.0 0.0" rot="0.0 0.0 0.0">
                                <vox pos="-0.025 -0.125 -0.05" rot="0.0 0.0 0.0" file="MOD/vox/physgun.vox" object="core_2" scale="0.5"/>
                            </group>
                            <group name="spine3" pos="0.0 0.0 -0.1" rot="0.0 0.0 0.0">
                                <vox pos="-0.025 -0.125 -0.05" rot="0.0 0.0 0.0" file="MOD/vox/physgun.vox" object="cannon" scale="0.5"/>
                                <location name="nozzle" pos="0.0 0.0 -0.15" rot="0.0 0.0 0.0"/>
                                <group name="arms_rot" pos="0.0 0.0 -0.05" rot="0.0 0.0 0.0">
                                    <group name="arm0_base" pos="0.0 0.1 0.0" rot="90.0 0.0 0.0">
                                        <vox pos="-0.025 0.025 0.05" rot="-90.0 0.0 0.0" file="MOD/vox/physgun.vox" object="arm_00" scale="0.5"/>
                                        <group name="arm0_tip" pos="0.0 0.0 -0.2" rot="0.0 0.0 0.0">
                                            <vox pos="-0.025 0.025 0.0" rot="-90.0 0.0 0.0" file="MOD/vox/physgun.vox" object="arm_01" scale="0.5"/>
                                        </group>
                                    </group>
                                    <group name="arm1_base" pos="-0.087 -0.05 0.0" rot="-90.0 180.0 60.0">
                                        <vox pos="-0.025 0.025 0.05" rot="-90.0 0.0 0.0" file="MOD/vox/physgun.vox" object="arm_10" scale="0.5"/>
                                        <group name="arm1_tip" pos="0.0 0.0 -0.2" rot="0.0 0.0 0.0">
                                            <vox pos="-0.025 0.025 0.0" rot="-90.0 0.0 0.0" file="MOD/vox/physgun.vox" object="arm_11" scale="0.5"/>
                                        </group>
                                    </group>
                                    <group name="arm2_base" pos="0.087 -0.05 0.0" rot="-90.0 -180.0 -60.0">
                                        <vox pos="-0.025 0.025 0.05" rot="-90.0 0.0 0.0" file="MOD/vox/physgun.vox" object="arm_20" scale="0.5"/>
                                        <group name="arm2_tip" pos="0.0 0.0 -0.2" rot="0.0 0.0 0.0">
                                            <vox pos="-0.025 0.025 0.0" rot="-90.0 0.0 0.0" file="MOD/vox/physgun.vox" object="arm_21" scale="0.5"/>
                                        </group>
                                    </group>
                                </group>
                            </group>
                        </group>
                    </group>
                </group>
            </group>
        </prefab>
	]],
	objects = {
		{ "cannon", Vec( 5, 3, 5 ) },
		{ "core_2", Vec( 5, 2, 5 ) },
		{ "core_1", Vec( 5, 2, 5 ) },
		{ "core_0", Vec( 5, 2, 5 ) },
		{ "arm_21", Vec( 1, 1, 2 ) },
		{ "arm_11", Vec( 1, 1, 2 ) },
		{ "arm_01", Vec( 1, 1, 2 ) },
		{ "arm_20", Vec( 1, 1, 5 ) },
		{ "arm_10", Vec( 1, 1, 5 ) },
		{ "arm_00", Vec( 1, 1, 5 ) },
		{ "body", Vec( 9, 6, 5 ) },
	},
}

RegisterToolUMF( "physgun", TOOL )
